@@ -13,6 +13,7 @@ from jinja2 import nodes
 from ..model import Scalar, Dictionary, List, Unknown, Tuple, String, Number, Boolean
 from ..mergers import merge_rtypes, merge, merge_many, merge_bool_expr_structs
 from ..exceptions import InvalidExpression, UnexpectedExpression, MergeException
+from ..config import default_config
 from .. import _compat
 from .util import visit_many
 
@@ -130,14 +131,14 @@ def visits_expr(node_cls):
     def decorator(func):
         expr_visitors[node_cls] = func
         @functools.wraps(func)
-        def wrapped_func(ast, ctx, macroses, config):
+        def wrapped_func(ast, ctx, macroses=None, config=default_config):
             assert isinstance(ast, node_cls)
-            return func(ast, ctx, macroses, config)
+            return func(ast, ctx, macroses=macroses, config=config)
         return wrapped_func
     return decorator
 
 
-def visit_expr(ast, ctx, macroses, config):
+def visit_expr(ast, ctx, macroses=None, config=default_config):
     """Returns a structure of ``ast``.
 
     :param ctx: :class:`Context`
@@ -152,10 +153,10 @@ def visit_expr(ast, ctx, macroses, config):
                 visitor = visitor_
     if not visitor:
         raise Exception('expression visitor for {0} is not found'.format(type(ast)))
-    return visitor(ast, ctx, macroses, config)
+    return visitor(ast, ctx, macroses, config=config)
 
 
-def _visit_dict(ast, ctx, macroses, items, config):
+def _visit_dict(ast, ctx, macroses, items, config=default_config):
     """A common logic behind nodes.Dict and nodes.Call (``{{ dict(a=1) }}``)
     visitors.
 
@@ -166,10 +167,10 @@ def _visit_dict(ast, ctx, macroses, items, config):
     struct = Dictionary()
     for key, value in items:
         value_rtype, value_struct = visit_expr(value, Context(
-            predicted_struct=Unknown.from_ast(value)), macroses, config)
+            predicted_struct=Unknown.from_ast(value)), macroses, config=config)
         struct = merge(struct, value_struct)
         if isinstance(key, nodes.Node):
-            key_rtype, key_struct = visit_expr(key, Context(predicted_struct=Scalar.from_ast(key)), macroses, config)
+            key_rtype, key_struct = visit_expr(key, Context(predicted_struct=Scalar.from_ast(key)), macroses, config=config)
             struct = merge(struct, key_struct)
             if isinstance(key, nodes.Const):
                 rtype[key.value] = value_rtype
@@ -179,32 +180,32 @@ def _visit_dict(ast, ctx, macroses, items, config):
 
 
 @visits_expr(nodes.BinExpr)
-def visit_bin_expr(ast, ctx, macroses, config):
-    l_rtype, l_struct = visit_expr(ast.left, ctx, macroses, config)
-    r_rtype, r_struct = visit_expr(ast.right, ctx, macroses, config)
+def visit_bin_expr(ast, ctx, macroses=None, config=default_config):
+    l_rtype, l_struct = visit_expr(ast.left, ctx, macroses, config=config)
+    r_rtype, r_struct = visit_expr(ast.right, ctx, macroses, config=config)
     rv = merge_bool_expr_structs(l_struct, r_struct)
     return merge_rtypes(l_rtype, r_rtype, operator=ast.operator), rv
 
 
 @visits_expr(nodes.UnaryExpr)
-def visit_unary_expr(ast, ctx, macroses, config):
-    return visit_expr(ast.node, ctx, macroses, config)
+def visit_unary_expr(ast, ctx, macroses=None, config=default_config):
+    return visit_expr(ast.node, ctx, macroses, config=config)
 
 
 @visits_expr(nodes.Compare)
-def visit_compare(ast, ctx, macroses, config):
+def visit_compare(ast, ctx, macroses=None, config=default_config):
     ctx.meet(Boolean(), ast)
     rtype, struct = visit_expr(ast.expr, Context(predicted_struct=Unknown.from_ast(ast.expr)),
-                               macroses, config)
+                               macroses, config=config)
     for op in ast.ops:
         op_rtype, op_struct = visit_expr(op.expr, Context(predicted_struct=Unknown.from_ast(ast.expr)),
-                                         macroses, config)
+                                         macroses, config=config)
         struct = merge(struct, op_struct)
     return Boolean.from_ast(ast), struct
 
 
 @visits_expr(nodes.Slice)
-def visit_slice(ast, ctx, macroses, config):
+def visit_slice(ast, ctx, macroses=None, config=default_config):
     nodes = [node for node in [ast.start, ast.stop, ast.step] if node is not None]
     struct = visit_many(nodes, macroses, config,
                         predicted_struct_cls=Number,
@@ -213,7 +214,7 @@ def visit_slice(ast, ctx, macroses, config):
 
 
 @visits_expr(nodes.Name)
-def visit_name(ast, ctx, macroses, config):
+def visit_name(ast, ctx, macroses=None, config=default_config):
     kwargs = {}
     return ctx.return_struct_cls.from_ast(ast, **kwargs), Dictionary({
         ast.name: ctx.get_predicted_struct(label=ast.name)
@@ -221,17 +222,17 @@ def visit_name(ast, ctx, macroses, config):
 
 
 @visits_expr(nodes.Getattr)
-def visit_getattr(ast, ctx, macroses, config):
+def visit_getattr(ast, ctx, macroses=None, config=default_config):
     context = Context(
         ctx=ctx,
         predicted_struct=Dictionary.from_ast(ast, {
             ast.attr: ctx.get_predicted_struct(label=ast.attr),
         }))
-    return visit_expr(ast.node, context, macroses, config)
+    return visit_expr(ast.node, context, macroses, config=config)
 
 
 @visits_expr(nodes.Getitem)
-def visit_getitem(ast, ctx, macroses, config):
+def visit_getitem(ast, ctx, macroses=None, config=default_config):
     arg = ast.arg
     if isinstance(arg, nodes.Const):
         if isinstance(arg.value, int):
@@ -260,15 +261,15 @@ def visit_getitem(ast, ctx, macroses, config):
         elif config.TYPE_OF_VARIABLE_INDEXED_WITH_VARIABLE_TYPE == 'dictionary':
             predicted_struct = Dictionary.from_ast(ast)
 
-    _, arg_struct = visit_expr(arg, Context(predicted_struct=Scalar.from_ast(arg)), macroses, config)
+    _, arg_struct = visit_expr(arg, Context(predicted_struct=Scalar.from_ast(arg)), macroses, config=config)
     rtype, struct = visit_expr(ast.node, Context(
         ctx=ctx,
-        predicted_struct=predicted_struct), macroses, config)
+        predicted_struct=predicted_struct), macroses, config=config)
     return rtype, merge(struct, arg_struct)
 
 
 @visits_expr(nodes.Test)
-def visit_test(ast, ctx, macroses, config):
+def visit_test(ast, ctx, macroses=None, config=default_config):
     ctx.meet(Boolean(), ast)
     if ast.name in ('divisibleby', 'escaped', 'even', 'lower', 'odd', 'upper'):
         # TODO
@@ -283,31 +284,31 @@ def visit_test(ast, ctx, macroses, config):
     else:
         raise InvalidExpression(ast, 'unknown test "{0}"'.format(ast.name))
     rtype, struct = visit_expr(ast.node, Context(return_struct_cls=Boolean,
-                                                 predicted_struct=predicted_struct), macroses, config)
+                                                 predicted_struct=predicted_struct), macroses, config=config)
     if ast.name == 'divisibleby':
         if not ast.args:
             raise InvalidExpression(ast, 'divisibleby must have an argument')
         _, arg_struct = visit_expr(ast.args[0],
-                                   Context(predicted_struct=Number.from_ast(ast.args[0])), macroses, config)
+                                   Context(predicted_struct=Number.from_ast(ast.args[0])), macroses, config=config)
         struct = merge(arg_struct, struct)
     return rtype, struct
 
 
 @visits_expr(nodes.Concat)
-def visit_concat(ast, ctx, macroses, config):
+def visit_concat(ast, ctx, macroses=None, config=default_config):
     ctx.meet(Scalar(), ast)
     return String.from_ast(ast), visit_many(ast.nodes, macroses, config, predicted_struct_cls=String)
 
 
 @visits_expr(nodes.CondExpr)
-def visit_cond_expr(ast, ctx, macroses, config):
-    if config.CONSIDER_CONDITIONS_AS_BOOLEAN:
+def visit_cond_expr(ast, ctx, macroses=None, config=default_config):
+    if config.BOOLEAN_CONDITIONS:
         test_predicted_struct = Boolean.from_ast(ast.test)
     else:
         test_predicted_struct = Unknown.from_ast(ast.test)
-    test_rtype, test_struct = visit_expr(ast.test, Context(predicted_struct=test_predicted_struct), macroses, config)
-    if_rtype, if_struct = visit_expr(ast.expr1, ctx, macroses, config)
-    else_rtype, else_struct = visit_expr(ast.expr2, ctx, macroses, config)
+    test_rtype, test_struct = visit_expr(ast.test, Context(predicted_struct=test_predicted_struct), macroses, config=config)
+    if_rtype, if_struct = visit_expr(ast.expr1, ctx, macroses, config=config)
+    else_rtype, else_struct = visit_expr(ast.expr2, ctx, macroses, config=config)
     struct = merge_many(if_struct, test_struct, else_struct)
     rtype = merge_rtypes(if_rtype, else_rtype)
 
@@ -319,148 +320,97 @@ def visit_cond_expr(ast, ctx, macroses, config):
                 lookup_struct = else_struct
             struct[var_name].may_be_defined = (lookup_struct and
                                                var_name in lookup_struct and
-                                               lookup_struct[var_name].assigned)
-        struct[var_name].checked_as_defined = test_struct[var_name].checked_as_defined
-        struct[var_name].checked_as_undefined = test_struct[var_name].checked_as_undefined
+                                               lookup_struct[var_name].constant)
+            struct[var_name].checked_as_defined = test_struct[var_name].checked_as_defined and (
+                not lookup_struct or not var_name in lookup_struct or lookup_struct[var_name].constant
+            )
+            struct[var_name].checked_as_undefined = test_struct[var_name].checked_as_undefined and (
+                not lookup_struct or not var_name in lookup_struct or lookup_struct[var_name].constant
+            )
 
     return rtype, struct
 
 
-def check_macro_call(ast, macro, args, kwargs):
-    # XXX the code needs to be refactored
-    rv = Dictionary()
-
-    expected_args = macro.args[:]
-    expected_kwargs = macro.kwargs[:]
-
-    def match_args(args, expected_args):
-        matched_args = list(zip(args, expected_args))
-        for i, (arg, (expected_arg_name, expected_arg)) in enumerate(matched_args, start=1):
-            old_label = arg.label
-            arg.label = 'argument #{0}'.format(i)
-            arg_struct = merge(expected_arg, arg)
-            if old_label:
-                arg_struct.label = old_label
-                rv[old_label] = arg_struct
-
-        args = args[len(matched_args):]
-        expected_args = expected_args[len(matched_args):]
-        return args, expected_args
-
-    args, expected_args = match_args(args, expected_args)
-
-    if args:
-        args, expected_kwargs = match_args(args, expected_kwargs)
-
-
-    def match_kwarg(kwarg_name, kwarg, expected_kwargs):
-        r = []
-        found = False
-        for (expected_kwarg_name, expected_kwarg) in expected_kwargs:
-            if kwarg_name == expected_kwarg_name:
-                kwarg_struct = merge(expected_kwarg, kwarg)  # to make sure
-                if kwarg.label:
-                    kwarg_struct.label = kwarg.label
-                    rv[kwarg.label] = kwarg_struct
-                found = True
-            else:
-                r.append((expected_kwarg_name, expected_kwarg))
-        return r, found
-
-    for kwarg, kwarg_type in _compat.iteritems(kwargs):
-        expected_args, found_1 = match_kwarg(kwarg, kwarg_type, expected_args)
-        expected_kwargs, found_2 = match_kwarg(kwarg, kwarg_type, expected_kwargs)
-        if not (found_1 or found_2):
-            raise InvalidExpression(ast, 'incorrect usage of "{0}". unknown argument "{1}"'.format(macro.name, kwarg))
-
-
-    if args or expected_args:
-        raise InvalidExpression(ast, ('incorrect usage of "{0}". it takes '
-                                      'exactly {1} positional arguments'.format(macro.name, len(macro.args))))
-
-    return rv
-
-
 @visits_expr(nodes.Call)
-def visit_call(ast, ctx, macroses, config):
+def visit_call(ast, ctx, macroses=None, config=default_config):
     if isinstance(ast.node, nodes.Name):
-        if ast.node.name in macroses:
-            # XXX the code needs to be refactored
-            struct = Dictionary()
+        if macroses and ast.node.name in macroses:
             macro = macroses[ast.node.name]
+            call = MacroCall(macro, ast.args, ast.kwargs, config=config)
+            args_struct = call.match_passed_args_to_expected_args()
+            if call.passed_args:
+                args_struct = merge(args_struct, call.match_passed_args_to_expected_kwargs())
+            if call.passed_kwargs:
+                args_struct = merge(args_struct, call.match_passed_kwargs_to_expected_args())
+            if call.passed_kwargs:
+                args_struct = merge(args_struct, call.match_passed_kwargs_to_expected_kwargs())
 
-            arg_structures = []
-            for arg_ast in ast.args:
-                arg_rtype, arg_struct = visit_expr(
-                    arg_ast, Context(predicted_struct=Unknown.from_ast(arg_ast)), macroses, config)
-                struct = merge(struct, arg_struct)
-                arg_structures.append(arg_rtype)
-
-            kwarg_structures = {}
-            for kwarg_ast in ast.kwargs:
-                kwarg_rtype, kwarg_struct = visit_expr(
-                    kwarg_ast.value, Context(predicted_struct=Unknown.from_ast(kwarg_ast)), macroses, config)
-                struct = merge(struct, kwarg_struct)
-                kwarg_structures[kwarg_ast.key] = kwarg_rtype
-
-            args_struct = check_macro_call(ast, macro, arg_structures, kwarg_structures)
-
-            return Unknown(), merge(args_struct, struct)
+            if call.passed_args or call.expected_args:
+                raise InvalidExpression(ast, ('incorrect usage of "{0}". it takes '
+                                              'exactly {1} positional arguments'.format(macro.name, len(macro.args))))
+            if call.passed_kwargs:
+                first_unknown_kwarg = next(_compat.iterkeys(call.passed_kwargs))
+                raise InvalidExpression(ast, ('incorrect usage of "{0}". unknown keyword argument '
+                                              '"{1}" is passed'.format(macro.name, first_unknown_kwarg)))
+            return Unknown(), args_struct
         elif ast.node.name == 'range':
             ctx.meet(List(Unknown()), ast)
             struct = Dictionary()
             for arg in ast.args:
                 arg_rtype, arg_struct = visit_expr(arg, Context(
-                    predicted_struct=Number.from_ast(arg)), macroses, config)
+                    predicted_struct=Number.from_ast(arg)), macroses, config=config)
                 struct = merge(struct, arg_struct)
             return List(Number()), struct
         elif ast.node.name == 'lipsum':
             ctx.meet(Scalar(), ast)
             struct = Dictionary()
-            # perhaps TODO: set possible types for args and kwargs
+            # probable TODO: set possible types for args and kwargs
             for arg in ast.args:
-                arg_rtype, arg_struct = visit_expr(arg, Context(predicted_struct=Scalar.from_ast(arg)), macroses, config)
+                arg_rtype, arg_struct = visit_expr(arg, Context(predicted_struct=Scalar.from_ast(arg)), macroses, config=config)
                 struct = merge(struct, arg_struct)
             for kwarg in ast.kwargs:
-                arg_rtype, arg_struct = visit_expr(kwarg.value, Context(predicted_struct=Scalar.from_ast(kwarg)), macroses, config)
+                arg_rtype, arg_struct = visit_expr(kwarg.value, Context(predicted_struct=Scalar.from_ast(kwarg)), macroses, config=config)
                 struct = merge(struct, arg_struct)
-            return Scalar(), struct
+            return String(), struct
         elif ast.node.name == 'dict':
             ctx.meet(Dictionary(), ast)
             if ast.args:
                 raise InvalidExpression(ast, 'dict accepts only keyword arguments')
-            return _visit_dict(ast, ctx, macroses, [(kwarg.key, kwarg.value) for kwarg in ast.kwargs], config)
+            return _visit_dict(ast, ctx, macroses, [(kwarg.key, kwarg.value) for kwarg in ast.kwargs], config=config)
         else:
             raise InvalidExpression(ast, '"{0}" call is not supported'.format(ast.node.name))
     elif isinstance(ast.node, nodes.Getattr):
         if ast.node.attr in ('keys', 'iterkeys', 'values', 'itervalues'):
             ctx.meet(List(Unknown()), ast)
             rtype, struct = visit_expr(
-                ast.node.node, Context(predicted_struct=Dictionary.from_ast(ast.node.node)), macroses, config)
+                ast.node.node, Context(predicted_struct=Dictionary.from_ast(ast.node.node)), macroses, config=config)
             return List(Unknown()), struct
         if ast.node.attr in ('startswith', 'endswith'):
             ctx.meet(Boolean(), ast)
             rtype, struct = visit_expr(
-                ast.node.node, Context(predicted_struct=String.from_ast(ast.node.node)), macroses, config)
+                ast.node.node, Context(predicted_struct=String.from_ast(ast.node.node)), macroses, config=config)
             return Boolean(), struct
         else:
             raise InvalidExpression(ast, '"{0}" call is not supported'.format(ast.node.attr))
 
 
 @visits_expr(nodes.Filter)
-def visit_filter(ast, ctx, macroses, config):
+def visit_filter(ast, ctx, macroses=None, config=default_config):
     return_struct_cls = None
     if ast.name in ('abs', 'striptags', 'capitalize', 'center', 'escape', 'filesizeformat',
                     'float', 'forceescape', 'format', 'indent', 'int', 'replace', 'round',
                     'safe', 'string', 'striptags', 'title', 'trim', 'truncate', 'upper',
                     'urlencode', 'urlize', 'wordcount', 'wordwrap', 'e'):
         ctx.meet(Scalar(), ast)
-        if ast.name in ('abs', 'float', 'int', 'round', ):
+        if ast.name in ('abs', 'round'):
             node_struct = Number.from_ast(ast.node)
             return_struct_cls = Number
+        elif ast.name in ('float', 'int'):
+            node_struct = Scalar.from_ast(ast.node)
+            return_struct_cls = Number
         elif ast.name in ('striptags', 'capitalize', 'center', 'escape', 'forceescape', 'format', 'indent',
                           'replace', 'safe', 'title', 'trim', 'truncate', 'upper', 'urlencode',
-                          'urliize', 'wordwrap', 'e'):
+                          'urlize', 'wordwrap', 'e'):
             node_struct = String.from_ast(ast.node)
             return_struct_cls = String
         elif ast.name == 'filesizeformat':
@@ -476,13 +426,17 @@ def visit_filter(ast, ctx, macroses, config):
             node_struct = Scalar.from_ast(ast.node)
     elif ast.name in ('batch', 'slice'):
         ctx.meet(List(List(Unknown())), ast)
-        node_struct = merge(
-            List(List(Unknown(), linenos=[ast.node.lineno]), linenos=[ast.node.lineno]),
-            ctx.get_predicted_struct()
-        ).item
+        rtype = List(List(Unknown(), linenos=[ast.node.lineno]), linenos=[ast.node.lineno])
+        node_struct = merge(rtype, ctx.get_predicted_struct()).item
+        _, struct = visit_expr(ast.node, Context(
+            ctx=ctx,
+            return_struct_cls=return_struct_cls,
+            predicted_struct=node_struct
+        ), macroses, config=config)
+        return rtype, struct
     elif ast.name == 'default':
         default_value_rtype, default_value_struct = visit_expr(
-            ast.args[0], Context(predicted_struct=Unknown.from_ast(ast.args[0])), macroses, config)
+            ast.args[0], Context(predicted_struct=Unknown.from_ast(ast.args[0])), macroses, config=config)
         node_struct = merge(
             ctx.get_predicted_struct(),
             default_value_rtype,
@@ -497,14 +451,16 @@ def visit_filter(ast, ctx, macroses, config):
         rtype, struct = visit_expr(ast.node, Context(
             return_struct_cls=String,
             predicted_struct=node_struct
-        ), macroses, config)
-        arg_rtype, arg_struct = visit_expr(ast.args[0], Context(predicted_struct=Scalar.from_ast(ast.args[0])), macroses, config)
+        ), macroses, config=config)
+        arg_rtype, arg_struct = visit_expr(ast.args[0],
+                                           Context(predicted_struct=String.from_ast(ast.args[0])), macroses, config=config)
         return rtype, merge(struct, arg_struct)
     elif ast.name in ('first', 'last', 'random', 'length', 'sum'):
         if ast.name in ('first', 'last', 'random'):
             el_struct = ctx.get_predicted_struct()
         elif ast.name == 'length':
             ctx.meet(Scalar(), ast)
+            return_struct_cls = Number
             el_struct = Unknown()
         else:
             ctx.meet(Scalar(), ast)
@@ -536,19 +492,19 @@ def visit_filter(ast, ctx, macroses, config):
         ctx=ctx,
         return_struct_cls=return_struct_cls,
         predicted_struct=node_struct
-    ), macroses, config)
+    ), macroses, config=config)
     return rv
 
 
 # :class:`nodes.Literal` visitors
 
 @visits_expr(nodes.TemplateData)
-def visit_template_data(ast, ctx, macroses, config):
+def visit_template_data(ast, ctx, macroses=None, config=default_config):
     return Scalar(), Dictionary()
 
 
 @visits_expr(nodes.Const)
-def visit_const(ast, ctx, macroses, config):
+def visit_const(ast, ctx, macroses=None, config=default_config):
     ctx.meet(Scalar(), ast)
     if isinstance(ast.value, _compat.string_types):
         rtype = String.from_ast(ast, constant=True)
@@ -562,13 +518,13 @@ def visit_const(ast, ctx, macroses, config):
 
 
 @visits_expr(nodes.Tuple)
-def visit_tuple(ast, ctx, macroses, config):
+def visit_tuple(ast, ctx, macroses=None, config=default_config):
     ctx.meet(Tuple(None), ast)
 
     struct = Dictionary()
     item_structs = []
     for item in ast.items:
-        item_rtype, item_struct = visit_expr(item, ctx, macroses, config)
+        item_rtype, item_struct = visit_expr(item, ctx, macroses, config=config)
         item_structs.append(item_rtype)
         struct = merge(struct, item_struct)
     rtype = Tuple.from_ast(ast, item_structs, constant=True)
@@ -576,14 +532,14 @@ def visit_tuple(ast, ctx, macroses, config):
 
 
 @visits_expr(nodes.List)
-def visit_list(ast, ctx, macroses, config):
+def visit_list(ast, ctx, macroses=None, config=default_config):
     ctx.meet(List(Unknown()), ast)
     struct = Dictionary()
 
     predicted_struct = merge(List(Unknown()), ctx.get_predicted_struct()).item
     el_rtype = None
     for item in ast.items:
-        item_rtype, item_struct = visit_expr(item, Context(predicted_struct=predicted_struct), macroses, config)
+        item_rtype, item_struct = visit_expr(item, Context(predicted_struct=predicted_struct), macroses, config=config)
         struct = merge(struct, item_struct)
         if el_rtype is None:
             el_rtype = item_rtype
@@ -594,6 +550,9 @@ def visit_list(ast, ctx, macroses, config):
 
 
 @visits_expr(nodes.Dict)
-def visit_dict(ast, ctx, macroses, config):
+def visit_dict(ast, ctx, macroses=None, config=default_config):
     ctx.meet(Dictionary(), ast)
-    return _visit_dict(ast, ctx, macroses, [(item.key, item.value) for item in ast.items], config)
+    return _visit_dict(ast, ctx, macroses, [(item.key, item.value) for item in ast.items], config=config)
+
+
+from ..macro import MacroCall
