@@ -9,9 +9,10 @@ Statement visitors return :class:`.models.Dictionary` of structures of variables
 import functools
 
 from jinja2 import nodes
-from jinja2schema.util import debug_repr
+from jinja2schema.config import default_config
 
-from ..model import Scalar, Dictionary, List, Unknown, Tuple, Boolean, Macro
+from ..model import Scalar, Dictionary, List, Unknown, Tuple, Boolean
+from ..macro import Macro
 from ..mergers import merge, merge_many
 from ..exceptions import InvalidExpression
 from .. import _compat
@@ -30,14 +31,14 @@ def visits_stmt(node_cls):
     def decorator(func):
         stmt_visitors[node_cls] = func
         @functools.wraps(func)
-        def wrapped_func(ast, macroses, config):
+        def wrapped_func(ast, macroses=None, config=default_config):
             assert isinstance(ast, node_cls)
             return func(ast, macroses, config)
         return wrapped_func
     return decorator
 
 
-def visit_stmt(ast, macroses, config):
+def visit_stmt(ast, macroses=None, config=default_config):
     """Returns a structure of ``ast``.
 
     :param ast: instance of :class:`jinja2.nodes.Stmt`
@@ -54,7 +55,7 @@ def visit_stmt(ast, macroses, config):
 
 
 @visits_stmt(nodes.For)
-def visit_for(ast, macroses, config):
+def visit_for(ast, macroses=None, config=default_config):
     body_struct = visit_many(ast.body, macroses, config, predicted_struct_cls=Scalar)
     else_struct = visit_many(ast.else_, macroses, config, predicted_struct_cls=Scalar)
 
@@ -83,8 +84,8 @@ def visit_for(ast, macroses, config):
 
 
 @visits_stmt(nodes.If)
-def visit_if(ast, macroses, config):
-    if config.CONSIDER_CONDITIONS_AS_BOOLEAN:
+def visit_if(ast, macroses=None, config=default_config):
+    if config.BOOLEAN_CONDITIONS:
         test_predicted_struct = Boolean.from_ast(ast.test)
     else:
         test_predicted_struct = Unknown.from_ast(ast.test)
@@ -102,14 +103,18 @@ def visit_if(ast, macroses, config):
                 lookup_struct = else_struct
             struct[var_name].may_be_defined = (lookup_struct and
                                                var_name in lookup_struct and
-                                               lookup_struct[var_name].assigned)
-        struct[var_name].checked_as_defined = test_struct[var_name].checked_as_defined
-        struct[var_name].checked_as_undefined = test_struct[var_name].checked_as_undefined
+                                               lookup_struct[var_name].constant)
+            struct[var_name].checked_as_defined = test_struct[var_name].checked_as_defined and (
+                not lookup_struct or not var_name in lookup_struct or lookup_struct[var_name].constant
+            )
+            struct[var_name].checked_as_undefined = test_struct[var_name].checked_as_undefined and (
+                not lookup_struct or not var_name in lookup_struct or lookup_struct[var_name].constant
+            )
     return struct
 
 
 @visits_stmt(nodes.Assign)
-def visit_assign(ast, macroses, config):
+def visit_assign(ast, macroses=None, config=default_config):
     struct = Dictionary()
     if (isinstance(ast.target, nodes.Name) or
             (isinstance(ast.target, nodes.Tuple) and isinstance(ast.node, nodes.Tuple))):
@@ -125,7 +130,6 @@ def visit_assign(ast, macroses, config):
         for var_name, var_ast in variables:
             var_rtype, var_struct = visit_expr(var_ast, Context(predicted_struct=Unknown.from_ast(var_ast)), macroses, config)
             var_rtype.constant = True
-            var_rtype.assigned = True
             var_rtype.label = var_name
             struct = merge_many(struct, var_struct, Dictionary({
                 var_name: var_rtype,
@@ -145,12 +149,12 @@ def visit_assign(ast, macroses, config):
 
 
 @visits_stmt(nodes.Output)
-def visit_output(ast, macroses, config):
+def visit_output(ast, macroses=None, config=default_config):
     return visit_many(ast.nodes, macroses, config, predicted_struct_cls=Scalar)
 
 
 @visits_stmt(nodes.Macro)
-def visit_macro(ast, macroses, config):
+def visit_macro(ast, macroses=None, config=default_config):
     # XXX the code needs to be refactored
     args = []
     kwargs = []
